import importlib
import pkgutil
import click


import pkgutil
import sys
import vendorless

package_cli = {}
for module_info in pkgutil.iter_modules(vendorless.__path__):
    try:
        mod = __import__(f'vendorless.{module_info.name}.commands', fromlist=['cli'])
        cli = getattr(mod, 'cli', None)
        if cli is not None:
            package_cli[module_info.name] = cli
    except ImportError:
        pass


@click.group()
def main():
    """Dispatcher CLI."""


# Dynamically add subcommands to the main group
for name, cmd in package_cli.items():
    main.add_command(cmd, name=name)


if __name__ == "__main__":
    main()




import yaml
import weakref
from dataclasses import dataclass, is_dataclass, asdict

_blueprints: weakref.WeakValueDictionary[int, 'Blueprint'] = weakref.WeakValueDictionary()

from typing import Generator, Iterator
from importlib.resources.abc import Traversable
from pathlib import PurePosixPath, Path
import importlib.resources

import jinja2

def get_template_dir_files(template_dir: Traversable, relative_to: PurePosixPath = PurePosixPath("")) -> Generator[str, None, None]:
    for child in template_dir.iterdir():
        rel_path = relative_to / child.name
        if child.is_dir():
            yield from get_template_dir_files(child, rel_path)
        else:
            yield str(rel_path)

from .templating import ResourceLoader

class Blueprint:
    def __init__(self) -> None:
        self._assert_is_dataclass()
    
    def _assert_is_dataclass(self):
        if not is_dataclass(self):
            raise TypeError(f"{self.__class__} must be a dataclass")

    def __post_init__(self):
        self._assert_is_dataclass()
        _blueprints[id(self)] = self

    def _copy_list(self) -> list[tuple[str, str]]:
        return []
    
    def _template_list(self) -> list[tuple[str, str]]:
        vl_module = self.__class__.__module__
        assert vl_module.startswith('vendorless.')
        assert len(vl_module.split('.')) >= 2

        template_dir = importlib.resources.files(f'{".".join(vl_module.split(".")[:2])}.templates')
        files = []
        for template_file in get_template_dir_files(template_dir):
            files.append((template_file, template_file))
        return files
    
    def _render(self, stack_root: Path, docker_compose: dict):
        loader = ResourceLoader(self)
        env = jinja2.Environment(
            loader=loader
        )
        context = asdict(self)

        dsts = {}
        copies = set()
        for src, dst in self._template_list():
            if dst in dsts:
                raise ValueError(f"multiple input files render '{dst}'")
            dsts[dst] = src
        
        for src, dst in self._copy_list():
            if dst in dsts:
                raise ValueError(f"multiple input files render '{dst}'")
            dsts[dst] = src
            copies.add(dst)

        for dst, src in dsts.items():
            dst = stack_root / env.from_string(dst).render(context)
            dst.parent.mkdir(parents=True, exist_ok=True)
            if dst in copies:
                with open(dst, 'wb') as f:
                    f.write((loader.files / src).read_bytes())  # TODO: test
            else:
                template = env.get_template(src)
                rendered = template.render(context)
                if dst.name == 'docker-compose.yaml':
                    
                    dc_data: dict = yaml.safe_load(rendered)
                    kv: dict
                    for first_key, kv in dc_data.items():
                        if first_key not in docker_compose:
                            docker_compose[first_key] = {}
                        
                        for second_key, v in kv.items():
                            if second_key in docker_compose[first_key]:
                                raise ValueError(f'multiple blueprints render {first_key}.{second_key} ')
                            docker_compose[first_key][second_key] = v
                else:
                    with open(dst, 'w') as f:
                        f.write(rendered)


    @classmethod
    def render_stack(cls, stack_root: str | Path):
        if isinstance(stack_root, str):
            stack_root = Path(stack_root)
        docker_compose = {}
        for blueprint in _blueprints.values():
            blueprint._render(stack_root, docker_compose)
        with open(stack_root/'docker-compose.yaml', 'w') as f:
            yaml.safe_dump(docker_compose, f)

@dataclass
class _DummyBlueprint(Blueprint):
    pass


import click
import importlib.resources
from cookiecutter.main import cookiecutter
import tempfile

from pathlib import Path

import runpy
import subprocess
from rich.console import Console
import shutil

import re
import os

from vendorless.core.blueprints import Blueprint

@click.group()
def cli():
    pass

@cli.command()
@click.option('-o', '--output-dir', type=click.Path(exists=True, file_okay=False, dir_okay=True), default='.', help='path to secrets dir')
def new(output_dir: str):
    """
    Create a new package. 
    """
    click.echo("Initializing new package.")
    templates_path = importlib.resources.files('vendorless.core.templates')
    cookiecutter(str(templates_path / 'package'), output_dir=output_dir)
    click.echo("New package initialized.")


@cli.group()
@click.argument('stack', type=click.STRING) # foo.py for local, package.module for package
@click.option('-s', '--secrets', type=click.Path(exists=True, file_okay=False, dir_okay=True), default=None, help='path to secrets dir')
def build(stack: str, secrets):
    """
    Build a stack.

    STACK is the module (.py file or package module) that defines the stack.
    """
    # if stack is local file -> build locally
    if stack.endswith('.py'):
        runpy.run_path(stack)
    else:
        runpy.run_module(stack)
    Blueprint.render_stack('output')


@cli.group()
@click.argument('stack', type=click.STRING)
def run(stack):
    """
    Run a stack.

    STACK is the module (.py file or package module) that defines the stack.
    """
    # if stack is local file -> build locally
    pass

@cli.group()
def dev():
    pass


def run_command(*command: str, return_stdout: bool=False, input: str=None, cwd=None, env=None) -> str:
    process = subprocess.Popen(
        command,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
        bufsize=1,
        cwd=cwd,
        env=env,
    )

    if not return_stdout:
        if input:
            process.stdin.write(input)
            process.stdin.flush()
            process.stdin.close()
            
        console = Console()
        with process.stdout:
            for line in iter(process.stdout.readline, ""):
                console.print(line, end="")
        process.wait()
        if process.returncode != 0:
            raise RuntimeError()
        return ""
    else:
        stdout, stderr = process.communicate()
        return stdout


@dev.command()
def docs_serve():
    run_command('mkdocs', 'serve')

@dev.command()
def docs_build():
    run_command('mkdocs', 'build', '-d', 'out/docs')


def extract_blocks(filepath: str, block: str):
    pattern = re.compile(fr"```{block} *\n(.*?)```", flags=re.MULTILINE | re.DOTALL)
    with open(filepath, 'r', encoding='utf-8') as f:
        matches = pattern.finditer(f.read())

    blocks = ''
    for match in matches:
        blocks += ''.join(match.groups())
    return blocks

@dev.command()
@click.argument('filepath', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option('-t', '--temp-dir', is_flag=True)
def docs_run(filepath: str, temp_dir: bool):

    bash_script = extract_blocks(filepath=filepath, block="console")
    bash_script = ''.join(l.removeprefix("$").strip(' ') for l in bash_script.splitlines(keepends=True) if l.startswith("$"))


    input = extract_blocks(filepath=filepath, block="salt")
    input = ''.join(l.split(':', maxsplit=1)[1].strip(' ') for l in input.splitlines(keepends=True))

    tmpdir = tempfile.TemporaryDirectory(prefix='vendorless.core.', delete=not temp_dir)

    env = os.environ.copy()
    env.pop("VIRTUAL_ENV", None)  # don't modify the current environment

    if not temp_dir:
        run_command(
            'bash', '-c', f"set -x\n{bash_script}",
            input=input,
            env=env,
        )
    else:
        with tmpdir:
            run_command(
                'bash', '-c', f"set -x\n{bash_script}",
                input=input,
                cwd=tmpdir.name,
                env=env,
            )
            shutil.rmtree(tmpdir.name)
        
    




# install and run 

# @click.group()



import importlib.resources as resources
from jinja2 import Environment, BaseLoader, TemplateNotFound
import importlib.resources

class ResourceLoader(BaseLoader):
    def __init__(self, obj):
        vl_module = obj.__class__.__module__
        assert vl_module.startswith('vendorless.')
        assert len(vl_module.split('.')) >= 2
        self.files = importlib.resources.files(f'{".".join(vl_module.split(".")[:2])}.templates')

    def get_source(self, environment, template):
        # resource_path = f"{self.folder}/{template}"
        try:
            contents = (self.files / template).read_text()
        except FileNotFoundError:
            raise TemplateNotFound(template)

        return contents, str(template), lambda: True



from dataclasses import dataclass

import inspect
from typing import Any


@dataclass
class ParameterReference:
    obj: object
    param: Any

    def dereference(self):
        return self.param.__get__(self.obj, self.obj.__class__)

UNRESOLVED = object()

class Parameter:
    def __init__(self, default=UNRESOLVED) -> None:
        self.default = default

    def __set_name__(self, owner, name):
        self.attr_name = f'_{name}'

    def __get__(self, instance, owner):
        if instance is None:
            return self
        
        is_resolved = hasattr(instance, self.attr_name) and not isinstance(instance, ParameterReference)

        if not is_resolved:
            return ParameterReference(instance, self)
        
        value = getattr(instance, self.attr_name)
        if isinstance(value, ParameterReference):
            value = value.dereference()

        return value
    
    def __set__(self, instance, value):
        if value is self:
            value = self.default
        
        if value is not UNRESOLVED:
            setattr(instance, self.attr_name, value)
    
    def __repr__(self) -> str:
        return f"<parameter {self.attr_name}>"
    
def parameter(default=UNRESOLVED) -> Any:
    return Parameter(default)

class computed_parameter: # pylint: disable=invalid-name
    def __init__(self, func):
        self.func = func
        self.attr_name = f"_{func.__name__}"

        sig = inspect.signature(func)

        self.precursors = [
            name for name, param in sig.parameters.items()
            if param.kind in (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD)
        ]
        assert self.precursors[0] == 'self'
        self.precursors = self.precursors[1:]
    
    def __get__(self, instance, owner):
        if instance is None:
            return self
        args = list(getattr(instance, p) for p in self.precursors)
        if any(isinstance(a, ParameterReference) for a in args):
            return ParameterReference(instance, self)
        value = self.func(instance, *args)
        return value



from .parameters import parameter
from .blueprints import Blueprint
from dataclasses import dataclass


@dataclass
class Volume(Blueprint):
    """
    Represents a Docker volume configuration.
    """

    name: str = parameter()
    """The name of the Docker volume."""

    def _template_list(self) -> list[tuple[str, str]]:
        return [
            ('volume/docker-compose.yaml', 'docker-compose.yaml')
        ]